@@ -39,7 +39,7 @@
 #include "drm.h"
 #include "drm_fourcc.h"
 
-#include "libdrm.h"
+#include "libdrm_macros.h"
 #include "xf86drm.h"
 
 #include "buffers.h"
@@ -98,8 +98,8 @@ struct format_info {
 	const struct yuv_info yuv;
 };
 
-#define MAKE_RGB_INFO(rl, ro, bl, bo, gl, go, al, ao) \
-	.rgb = { { (rl), (ro) }, { (bl), (bo) }, { (gl), (go) }, { (al), (ao) } }
+#define MAKE_RGB_INFO(rl, ro, gl, go, bl, bo, al, ao) \
+	.rgb = { { (rl), (ro) }, { (gl), (go) }, { (bl), (bo) }, { (al), (ao) } }
 
 #define MAKE_YUV_INFO(order, xsub, ysub, chroma_stride) \
 	.yuv = { (order), (xsub), (ysub), (chroma_stride) }
@@ -1162,6 +1162,8 @@ bo_create(int fd, unsigned int format,
 	switch (format) {
 	case DRM_FORMAT_NV12:
 	case DRM_FORMAT_NV21:
+	case DRM_FORMAT_YUV420:
+	case DRM_FORMAT_YVU420:
 		virtual_height = height * 3 / 2;
 		break;
 
